@@ -9,7 +9,7 @@
 #
 package Redis;
 {
-  $Redis::VERSION = '1.973_01';
+  $Redis::VERSION = '1.973_02';
 }
 
 # ABSTRACT: Perl binding for Redis database
@@ -426,23 +426,21 @@ sub wait_for_messages {
       $s->add($sock);
 
       while ($s->can_read($timeout)) {      
-        my $has_stuff = __try_read_sock($sock);
+        my $has_stuff = $self->__try_read_sock($sock);
         # If the socket is ready to read but there is nothing to read, ( so
         # it's an EOF ), try to reconnect.
         defined $has_stuff
           or $self->__throw_reconnect('EOF from server');
-        while (1) {
-          my $has_stuff = __try_read_sock($sock);
-          $has_stuff
-            or last ; ## no data ( or socket became EOF), back to select until
-                      ## timeout
-          do {
-              my ($reply, $error) = $self->__read_response('WAIT_FOR_MESSAGES');
-              croak "[WAIT_FOR_MESSAGES] $error, " if defined $error;
-              $self->__process_pubsub_msg($reply);
-              $count++;
-          } while ($self->{__buf});
-        }
+
+        do {
+          my ($reply, $error) = $self->__read_response('WAIT_FOR_MESSAGES');
+          croak "[WAIT_FOR_MESSAGES] $error, " if defined $error;
+          $self->__process_pubsub_msg($reply);
+          $count++;
+
+          # if __try_read_sock() return 0 (no data)
+          # or undef ( socket became EOF), back to select until timeout
+        } while ($self->{__buf} || $self->__try_read_sock($sock));
       }
     
     });
@@ -687,7 +685,7 @@ sub __send_command {
   }
 
   ## Check to see if socket was closed: reconnect on EOF
-  my $status = __try_read_sock($sock);
+  my $status = $self->__try_read_sock($sock);
   $self->__throw_reconnect('Not connected to any server')
     unless defined $status;
 
@@ -816,16 +814,30 @@ sub __read_len {
 }
 
 sub __try_read_sock {
-  my $sock = shift;
+  my ($self, $sock) = @_;
   my $data = '';
 
   while (1) {
-      my $res = recv($sock, $data, 1, MSG_PEEK | MSG_DONTWAIT);
-      my $err = 0 + $!;
+      # WIN32 doesn't support MSG_DONTWAIT,
+      # need to swith fh to nonblockng mode manually.
+      # For Unix still use MSG_DONTWAIT because of fewer syscalls
+      my ($res, $err);
+      if (WIN32) {
+          __fh_nonblocking_win32($sock, 1);
+          $res = recv($sock, $data, BUFSIZE, 0);
+          $err = 0 + $!;
+          __fh_nonblocking_win32($sock, 0);
+      } else {
+          $res = recv($sock, $data, BUFSIZE, MSG_DONTWAIT);
+          $err = 0 + $!;
+      }
 
       if (defined $res) {
-        ## have data
-        length($data) and return 1;
+        ## have read some data
+        if (length($data)) {
+            $self->{__buf} .= $data;
+            return 1;
+        }
 
         ## no data but also no error means EOF
         return;
@@ -844,6 +856,11 @@ sub __try_read_sock {
   }
 }
 
+## Copied from AnyEvent::Util
+sub __fh_nonblocking_win32 {
+    ioctl $_[0], 0x8004667e, pack "L", $_[1];
+}
+
 ##########################
 # I take exception to that
 
@@ -868,7 +885,7 @@ Redis - Perl binding for Redis database
 
 =head1 VERSION
 
-version 1.973_01
+version 1.973_02
 
 =head1 SYNOPSIS
 
